from typing import (
    Union,
)

from Crypto.Hash import (
    keccak,
)

from platon_hash.abc import (
    BackendAPI,
    PreImageAPI,
)


class CryptodomePreimage(PreImageAPI):

    def __init__(self, prehash: bytes) -> None:
        self._hash = keccak.new(data=prehash, digest_bits=256, update_after_digest=True)
        # pycryptodome doesn't expose a `copy` mechanism for it's hash objects
        # so we keep a record of all of the parts for when/if we need to copy
        # them.
        self._parts = [prehash]

    def update(self, prehash: bytes) -> None:
        self._hash.update(prehash)
        self._parts.append(prehash)

    def digest(self) -> bytes:
        return self._hash.digest()

    def copy(self) -> 'CryptodomePreimage':
        return CryptodomePreimage(b''.join(self._parts))


class CryptodomeBackend(BackendAPI):
    def keccak256(self, prehash: Union[bytearray, bytes]) -> bytes:
        hasher = keccak.new(data=prehash, digest_bits=256)
        return hasher.digest()

    def preimage(self, prehash: Union[bytearray, bytes]) -> PreImageAPI:
        return CryptodomePreimage(prehash)


backend = CryptodomeBackend()
keccak256 = backend.keccak256
preimage = backend.preimage



import pytest


@pytest.fixture(params=['auto', 'explicit', 'env'])
def keccak(monkeypatch, request, keccak_auto):
    if request.param == 'auto':
        return keccak_auto
    elif request.param == 'explicit':
        from platon_hash.backends import pysha3
        from platon_hash import Keccak256
        return Keccak256(pysha3)
    elif request.param == 'env':
        monkeypatch.setenv('platon_hash_BACKEND', 'pysha3')
        from platon_hash.auto import keccak
        return keccak
    else:
        raise AssertionError("Unrecognized approach to import keccak: %s" % request.param)



from typing import (
    Union,
)

from sha3 import (
    keccak_256 as _keccak_256,
)

from platon_hash.abc import (
    BackendAPI,
    PreImageAPI,
)


class Pysha3Preimage(PreImageAPI):

    def __init__(self, prehash: bytes) -> None:
        self._hash = _keccak_256(prehash)

    def update(self, prehash: bytes) -> None:
        return self._hash.update(prehash)  # type: ignore

    def digest(self) -> bytes:
        return self._hash.digest()  # type: ignore

    def copy(self) -> 'Pysha3Preimage':
        dup = Pysha3Preimage(b'')
        dup._hash = self._hash.copy()
        return dup


class PySha3Backend(BackendAPI):
    def keccak256(self, prehash: Union[bytearray, bytes]) -> bytes:
        return _keccak_256(prehash).digest()  # type: ignore

    def preimage(self, prehash: Union[bytearray, bytes]) -> PreImageAPI:
        return Pysha3Preimage(prehash)


backend = PySha3Backend()
keccak256 = backend.keccak256
preimage = backend.preimage



import pickle

import pytest

# Note that this file is symlink'd in all backend folders, so a change here is
# automatically shared across backend tests. If you want to add a backend-specific
# test, add it to a new file.


@pytest.fixture
def keccak_auto():
    from platon_hash.auto import keccak
    return keccak


@pytest.mark.parametrize(
    'prehash, expected_result',
    (
        (
            b'',
            b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';\x7b\xfa\xd8\x04]\x85\xa4p",  # noqa: E501
        ),
        (
            bytearray(b''),
            b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';\x7b\xfa\xd8\x04]\x85\xa4p",  # noqa: E501
        ),
    ),
)
def test_keccak_256(keccak, prehash, expected_result):
    assert keccak(prehash) == expected_result


@pytest.mark.parametrize(
    'parts, expected_result',
    (
        (
            [b''],
            b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';\x7b\xfa\xd8\x04]\x85\xa4p",  # noqa: E501
        ),
        (
            [bytearray(b'')],
            b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';\x7b\xfa\xd8\x04]\x85\xa4p",  # noqa: E501
        ),
        (
            [b'', b'', b''],
            b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';\x7b\xfa\xd8\x04]\x85\xa4p",  # noqa: E501
        ),
        (
            [bytearray(b''), bytearray(b''), bytearray(b'')],
            b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';\x7b\xfa\xd8\x04]\x85\xa4p",  # noqa: E501
        ),
        (
            [b'arst', b'tsra'],
            b"\xb1\xf3T\xb2\x8f\xf2\x84R\xd6\xb9\xd6\x1fA\x06\x1b\xbe\x82\xbe\xb1\xfc\x98\xf33d\xa8\x05\x8d\x1a]\x16M\x05",  # noqa: E501
        ),
        (
            [bytearray(b'arst'), bytearray(b'tsra')],
            b"\xb1\xf3T\xb2\x8f\xf2\x84R\xd6\xb9\xd6\x1fA\x06\x1b\xbe\x82\xbe\xb1\xfc\x98\xf33d\xa8\x05\x8d\x1a]\x16M\x05",  # noqa: E501
        ),
        (
            [bytearray(b'arst'), b'tsra'],
            b"\xb1\xf3T\xb2\x8f\xf2\x84R\xd6\xb9\xd6\x1fA\x06\x1b\xbe\x82\xbe\xb1\xfc\x98\xf33d\xa8\x05\x8d\x1a]\x16M\x05",  # noqa: E501
        ),
        (
            [b'arst', bytearray(b'tsra')],
            b"\xb1\xf3T\xb2\x8f\xf2\x84R\xd6\xb9\xd6\x1fA\x06\x1b\xbe\x82\xbe\xb1\xfc\x98\xf33d\xa8\x05\x8d\x1a]\x16M\x05",  # noqa: E501
        ),
    ),
)
def test_keccak_256_preimage(keccak, parts, expected_result):
    preimage = keccak.new(parts[0])
    for part in parts[1:]:
        preimage.update(part)
    assert preimage.digest() == expected_result


def test_copy_keccak_256_preimage(keccak):
    preimage_origin = keccak.new(b'')
    preimage_copy = preimage_origin.copy()

    preimage_origin.update(b'arsttsra')

    assert preimage_origin.digest() == b"\xb1\xf3T\xb2\x8f\xf2\x84R\xd6\xb9\xd6\x1fA\x06\x1b\xbe\x82\xbe\xb1\xfc\x98\xf33d\xa8\x05\x8d\x1a]\x16M\x05"  # noqa: E501
    assert preimage_copy.digest() == b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';\x7b\xfa\xd8\x04]\x85\xa4p"  # noqa: E501


def test_update_digest_update(keccak):
    preimage = keccak.new(b'')
    preimage.update(b'arst')
    assert preimage.digest() == b"\xc30[\xc9\xde\x12D\xe4\x80P\x96,\xedP\xb7Y4\xc3p\x06\xe9\x9e\x8bzb!>\x94\\=\xfc\xd7"  # noqa: E501
    preimage.update(b'tsra')
    assert preimage.digest() == b"\xb1\xf3T\xb2\x8f\xf2\x84R\xd6\xb9\xd6\x1fA\x06\x1b\xbe\x82\xbe\xb1\xfc\x98\xf33d\xa8\x05\x8d\x1a]\x16M\x05"  # noqa: E501


def test_auto_pickleable(keccak_auto):
    serialized = pickle.dumps(keccak_auto)
    deserialized = pickle.loads(serialized)

    assert deserialized(b'arst') == keccak_auto(b'arst')



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Union,
)


class PreImageAPI(ABC):
    @abstractmethod
    def __init__(self, value: bytes) -> None:
        ...

    @abstractmethod
    def update(self, value: bytes) -> None:
        ...

    @abstractmethod
    def digest(self) -> bytes:
        ...

    @abstractmethod
    def copy(self) -> 'PreImageAPI':
        ...


class BackendAPI(ABC):
    @abstractmethod
    def keccak256(self, in_data: Union[bytearray, bytes]) -> bytes:
        ...

    @abstractmethod
    def preimage(self, in_data: Union[bytearray, bytes]) -> PreImageAPI:
        ...



from typing import (
    Union,
)

from .abc import (
    BackendAPI,
    PreImageAPI,
)


class Keccak256:
    def __init__(self, backend: BackendAPI) -> None:
        self._backend = backend
        self.hasher = self._hasher_first_run
        self.preimage = self._preimage_first_run

    def _hasher_first_run(self, in_data: Union[bytearray, bytes]) -> bytes:
        """
        Validate, on first-run, that the hasher backend is valid.

        After first run, replace this with the new hasher method.
        This is a bit of a hacky way to minimize overhead on hash calls after
        this first one.
        """
        new_hasher = self._backend.keccak256
        assert new_hasher(b'') == b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';\x7b\xfa\xd8\x04]\x85\xa4p"  # noqa: E501
        self.hasher = new_hasher
        return new_hasher(in_data)

    def _preimage_first_run(self, in_data: Union[bytearray, bytes]) -> PreImageAPI:
        new_preimage = self._backend.preimage
        self.preimage = new_preimage
        return new_preimage(in_data)

    def __call__(self, preimage: Union[bytearray, bytes]) -> bytes:
        if not isinstance(preimage, (bytearray, bytes)):
            raise TypeError(
                "Can only compute the hash of `bytes` or `bytearray` values, not %r" % preimage
            )

        return self.hasher(preimage)

    def new(self, preimage: Union[bytearray, bytes]) -> PreImageAPI:
        if not isinstance(preimage, (bytearray, bytes)):
            raise TypeError(
                "Can only compute the hash of `bytes` or `bytearray` values, not %r" % preimage
            )
        return self.preimage(preimage)



import sys
from unittest import (
    mock,
)

import pytest


def clean_module(name):
    """Clean a module import so all variables in it will be created again"""
    try:
        del sys.modules[name]
    except KeyError:
        pass
    return


def test_import_auto():
    clean_module("platon_hash.auto")
    from platon_hash.auto import keccak  # noqa: F401


def test_import_auto_empty_crash(monkeypatch):
    clean_module("platon_hash.auto")
    from platon_hash.auto import keccak
    with mock.patch.dict("sys.modules", {"sha3": None, "Crypto.Hash": None}):
        with pytest.raises(ImportError, match="None of these hashing backends are installed"):
            keccak(b'eh')


def test_import():
    clean_module("platon_hash")
    import platon_hash  # noqa: F401


@pytest.mark.parametrize(
    'backend',
    [
        'pycryptodome',
        'pysha3',
    ],
)
def test_load_by_env(monkeypatch, backend):
    clean_module("platon_hash.auto")
    from platon_hash.auto import keccak
    monkeypatch.setenv('platon_hash_BACKEND', backend)
    with mock.patch.dict("sys.modules", {"sha3": None, "Crypto.Hash": None}):
        with pytest.raises(ImportError) as excinfo:
            keccak(b'triggered')
    expected_msg = (
        "The backend specified in platon_hash_BACKEND, '{0}', is not installed. "
        "Install with `pip install platon-hash[{0}]`.".format(backend)
    )
    assert expected_msg in str(excinfo.value)


def test_load_unavailable_backend_by_env(monkeypatch):
    clean_module("platon_hash.auto")
    from platon_hash.auto import keccak
    backend = 'this-backend-will-never-exist'
    monkeypatch.setenv('platon_hash_BACKEND', backend)
    with pytest.raises(ValueError) as excinfo:
        keccak(b'triggered')
    expected_msg = (
        "The backend specified in platon_hash_BACKEND, '{0}', is not supported. "
        "Choose one of".format(backend)
    )
    assert expected_msg in str(excinfo.value)



import importlib
import logging
import os

from platon_hash.abc import (
    BackendAPI,
)
from platon_hash.backends import (
    SUPPORTED_BACKENDS,
)


def auto_choose_backend() -> BackendAPI:
    env_backend = get_backend_in_environment()

    if env_backend:
        return load_environment_backend(env_backend)
    else:
        return choose_available_backend()


def get_backend_in_environment() -> str:
    return os.environ.get('platon_hash_BACKEND', "")


def load_backend(backend_name: str) -> BackendAPI:
    import_path = 'platon_hash.backends.%s' % backend_name
    module = importlib.import_module(import_path)

    try:
        backend = module.backend  # type: ignore
    except AttributeError as e:
        raise ValueError(
            "Import of %s failed, because %r does not have 'backend' attribute" % (
                import_path,
                module,
            )
        ) from e

    if isinstance(backend, BackendAPI):
        return backend
    else:
        raise ValueError(
            "Import of %s failed, because %r is an invalid back end" % (
                import_path,
                backend,
            )
        )


def load_environment_backend(env_backend: str) -> BackendAPI:
    if env_backend in SUPPORTED_BACKENDS:
        try:
            return load_backend(env_backend)
        except ImportError as e:
            raise ImportError(
                "The backend specified in platon_hash_BACKEND, '{0}', is not installed. "
                "Install with `pip install platon-hash[{0}]`.".format(env_backend)
            ) from e
    else:
        raise ValueError(
            "The backend specified in platon_hash_BACKEND, %r, is not supported. "
            "Choose one of: %r" % (env_backend, SUPPORTED_BACKENDS)
        )


def choose_available_backend() -> BackendAPI:
    for backend in SUPPORTED_BACKENDS:
        try:
            return load_backend(backend)
        except ImportError:
            logging.getLogger('platon_hash').debug("Failed to import %s", backend, exc_info=True)
    raise ImportError(
        "None of these hashing backends are installed: %r.\n"
        "Install with `pip install platon-hash[%s]`." % (
            SUPPORTED_BACKENDS,
            SUPPORTED_BACKENDS[0],
        )
    )



from typing import (
    Union,
)

from platon_hash.abc import (
    BackendAPI,
    PreImageAPI,
)
from platon_hash.utils import (
    auto_choose_backend,
)


class AutoBackend(BackendAPI):

    def _initialize(self) -> None:
        backend = auto_choose_backend()
        # Use setattr to circumvent mypy's confusion, see:
        # https://github.com/python/mypy/issues/2427
        setattr(self, 'keccak256', backend.keccak256)
        setattr(self, 'preimage', backend.preimage)

    def keccak256(self, in_data: Union[bytearray, bytes]) -> bytes:
        self._initialize()
        return self.keccak256(in_data)

    def preimage(self, in_data: Union[bytearray, bytes]) -> PreImageAPI:
        self._initialize()
        return self.preimage(in_data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    setup,
    find_packages,
)

extras_require = {
    'test': [
        "pytest==5.4.1",
        "pytest-xdist",
        "tox==3.14.6",
    ],
    'lint': [
        "flake8==3.7.9",
        "isort>=4.2.15,<5",
        "mypy==0.770",
        "pydocstyle>=5.0.0,<6",
    ],
    'doc': [
        "Sphinx>=1.6.5,<2",
        "sphinx_rtd_theme>=0.1.9,<1",
        "towncrier>=19.2.0, <20",
    ],
    'dev': [
        "bumpversion>=0.5.3,<1",
        "pytest-watch>=4.1.0,<5",
        "wheel",
        "twine",
        "ipython",
    ],
    # optional backends:
    'pycryptodome': [
        "pycryptodome>=3.6.6,<4",
    ],
    'pysha3': [
        "pysha3>=1.0.0,<2.0.0",
    ],
}

extras_require['dev'] = (
    extras_require['dev'] +  # noqa: W504
    extras_require['test'] +  # noqa: W504
    extras_require['lint'] +  # noqa: W504
    extras_require['doc']
)


with open('./README.md') as readme:
    long_description = readme.read()


setup(
    name='platon-hash',
    # *IMPORTANT*: Don't manually change the version here. Use `make bump`, as described in readme
    version='1.2.1',
    description="""platon-hash: The Platon hashing function, keccak256, sometimes (erroneously) called sha3""",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Shinnng',
    author_email='shinnng@outlook.com',
    url='https://github.com/platonnetwork/platon-hash',
    include_package_data=True,
    install_requires=[],
    python_requires='>=3.5, <4',
    extras_require=extras_require,
    py_modules=['platon_hash'],
    license="MIT",
    zip_safe=False,
    keywords='platon',
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={'platon_hash': ['py.typed']},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)



"""
A collection of optional backends that implement hashing.

You must manually select and install the backend you want. If the backend is
not installed, then trying to import the module for that backend will cause an
:class:`ImportError`.

See :ref:`Choose a hashing backend` for more.
"""

SUPPORTED_BACKENDS = [
    'pycryptodome',  # prefer this over pysha3, for pypy3 support
    'pysha3',
]
